/**
 * This code accompanies the paper:
 *
 * "A Simple and Effective Relevance-Based Point Sampling for 3D Shapes"
 * E. Rodola, A. Albarelli, D. Cremers, A. Torsello
 * Pattern Recognition Letters, 59(1), pp 41-47, 2015
 *
 * Please cite the paper above if you use this code.
 *
 * Written by Emanuele Rodola.
 **/
#include "mex.h"
#include <cmath>
#include <queue>
#include <iostream>
#include "mesh.h"

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray*prhs[])
{
	if (nrhs != 6 || nlhs != 1)
      mexErrMsgTxt("Usage: areas = calc_relevance(vertices, triangles, normals, areas, dot_thresh, max_dist).");

	const double* const pts = mxGetPr(prhs[0]);
	const int np = int( mxGetN(prhs[0]) );
	const double* const tri = mxGetPr(prhs[1]);
	const int nt = int( mxGetN(prhs[1]) );
	const double* const nrmls = mxGetPr(prhs[2]);
	const double* const areas = mxGetPr(prhs[3]);
	const double dot_thresh = *mxGetPr(prhs[4]);
	const double max_dist = *mxGetPr(prhs[5]);
	const double max_dist2 = max_dist*max_dist;
	
	if (np == 3)
		mexErrMsgTxt("It seems like you only have 3 vertices. Please try to transpose the input matrix.");
		
	if (nt == 3)
		mexErrMsgTxt("It seems like you only have 3 triangles. Please try to transpose the input matrix.");
		
    if (np != int( mxGetN(prhs[2]) ))
		mexErrMsgTxt("Each point must be given a normal.");
		
	if (int( mxGetM(prhs[2]) ) != 3)
		mexErrMsgTxt("It seems like you only have 3 normals. Please try to transpose the input matrix.");

	//std::cout << np << " vertices, " << nt << " triangles" << std::endl;
	
	// Load the mesh

	mesh_t mesh;
	
	std::vector< vec3d<double> > vertices(np);
	
	for (int i=0; i<np; ++i)
	{
		vec3d<double>& pt = vertices[i];
		pt.x = pts[i*3];
		pt.y = pts[i*3+1];
		pt.z = pts[i*3+2];
		//std::cout << pt << std::endl;
	}
	
	std::vector< vec3d<double> > normals(np);
	
	for (int i=0; i<np; ++i)
	{
		vec3d<double>& n = normals[i];
		n.x = nrmls[i*3];
		n.y = nrmls[i*3+1];
		n.z = nrmls[i*3+2];
	}
	
	mesh.put_vertices(vertices);
	
	for (int i=0; i<nt; ++i)
	{
		int a, b, c;
		a = tri[i*3] - 1; // 1-based to 0-based
		b = tri[i*3+1] - 1;
		c = tri[i*3+2] - 1;
		mesh.add_triangle(a,b,c);
		//std::cout << a << " " << b << " " << c << std::endl;
	}
	
	//
	
	plhs[0] = mxCreateDoubleMatrix(1, np, mxREAL);
	double* scores = mxGetPr(plhs[0]);

	for (int i=0; i<np; ++i)
		scores[i] = 0.;
	
	for (int i=0; i<np; ++i)
	{
		if (i % 1000 == 0) 
			std::cout << 100.*i/(double)np << "%" << std::endl;
		
		const vec3d<double>& normal_ref = normals[i];
		const vec3d<double>& pt_ref = vertices[i];
		
		std::vector<int> marked(np,false);
		std::queue<int> q;
		
		q.push(i);
		marked[i] = true;
		
		std::vector<size_t> neighs;

		int iter=0;
		while (!q.empty())
		{
			++iter;
			const int seed = q.front();
			q.pop();

			scores[i] += areas[seed];
			mesh.get_vertex_ring(seed, 1, neighs);

			for (size_t k=0; k<neighs.size(); ++k)
			{
				const int neigh = neighs[k];

				if (marked[neigh])
					continue;

				marked[neigh] = true;
				
				if (squared_dist(pt_ref, vertices[neigh]) > max_dist2)
					continue;

				const vec3d<double>& normal_neigh = normals[neigh];

				if (dot_product(normal_ref, normal_neigh) > dot_thresh)
				{
					//marked[neigh] = true;
					q.push(neigh);
				}
			}
		}
	}
}
